#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2023-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""Smart SSH Client."""

import argparse
import os
import os.path
import pathlib
import socket
import subprocess  # nosec
import sys
import urllib.parse
from typing import Any

from smart_ssh_client.ssh import add_scp_parser, add_sftp_parser, add_ssh_parser

from .config import get_host_config
from .exceptions import SmartSSHClientError
from .proxy import run_proxy
from .utils import get_temp_directory
from .version import __version__

__all__ = [
    "run_via_proxy",
    "main",
]


def run_via_proxy(  # noqa: C901, PLR0912, PLR0915
    command: str,
    args_dict: dict[str, Any],
    ssh_args: list[str],
) -> int:
    """Run SSH with ourselves as a proxy."""

    port = None
    port_knock = None
    username = None
    username_component = None
    destination = ""

    # Work out what the command line argument for the port is
    if command == "ssh":
        port_flag = "ssh_p"
        # Work out destination
        destination = args_dict["DESTINATION"]

    elif command == "scp":
        port_flag = "scp_P"
        # Check for : in the first argument
        if ":" in args_dict["SOURCES.. DESTINATION"][0]:
            destination = args_dict["SOURCES.. DESTINATION"][0].split(":")[0]
        # If not found check the second parameter
        elif len(args_dict["SOURCES.. DESTINATION"]) > 1 and ":" in args_dict["SOURCES.. DESTINATION"][1]:
            destination = args_dict["SOURCES.. DESTINATION"][1].split(":")[0]

    elif command == "sftp":
        port_flag = "sftp_P"
        # Work out destination
        destination = args_dict["DESTINATION"]

    else:
        raise SmartSSHClientError(f"Unknown command: {command}")

    # If there is a URI within the hostname, split off the destination_hostname as the first part
    destination_args = {}
    if destination.startswith(("ssh://", "scp://", "sftp://")):
        destination_parsed = urllib.parse.urlparse(destination)
        destination_args = urllib.parse.parse_qs(destination_parsed.query)
        # Check URL scheme
        if destination_parsed.scheme not in ["ssh", "scp", "sftp"]:
            raise SmartSSHClientError(f"Unknown URL scheme: {destination_parsed.scheme}")
        destination_hostname = destination_parsed.netloc
    else:
        destination_hostname = destination

    # Store the original destination_hostname so we can see if it was changed below and do a replace
    original_destination_hostname = destination_hostname

    # Split off username_component and hostname
    if "@" in destination_hostname:
        username_component, destination_hostname = destination_hostname.split("@", 1)

    # Work out port to use for lookup
    if port_flag and args_dict.get(port_flag):
        port = args_dict[port_flag]

    # Grab host config
    host_config = get_host_config(args_dict, destination_hostname, port)
    if host_config.get("hostname"):
        destination_hostname = host_config["hostname"]
    if host_config.get("port"):
        port = host_config["port"]
    if host_config.get("portknock"):
        port_knock = host_config["portknock"]
    if host_config.get("username"):
        username = host_config["username"]
    # Check for query parameters
    if destination_args.get("hostname"):
        destination_hostname = destination_args["hostname"][0]
    if destination_args.get("port"):
        port = destination_args["port"][0]
    if destination_args.get("portknock"):
        port_knock = destination_args["portknock"][0]
    # Check for command line overrides
    if args_dict.get("smart_ssh_knock"):
        port_knock = args_dict["smart_ssh_knock"]
    if port_flag and args_dict.get(port_flag):
        port = args_dict[port_flag]
    if username_component:
        username = username_component

    # Replace destination in ssh_args with destination without the query params
    for idx, ssh_arg in enumerate(ssh_args):
        if ssh_arg == destination:
            # Work out the new destination by replacing the original if it was changed
            new_destination = destination.split("?", maxsplit=1)[0]
            if original_destination_hostname != destination_hostname:
                new_destination = new_destination.replace(original_destination_hostname, destination_hostname)
            # Replace the SSH argument with the new destination
            ssh_args[idx] = new_destination

    # Work out our temp directory
    temp_dir = get_temp_directory()

    # SSH command with ProxyCommand pointing to this script
    ssh_command = [f"/usr/bin/{command}"]

    # Skip adding p11-kit-proxy.so if we have -I already specified
    if "-I" in ssh_args:
        ssh_command.extend(["-I", "/usr/lib/p11-kit-proxy.so"])

    # Set TCP keepalive options to keep our session open
    ssh_command.extend(["-o", "TCPKeepAlive=yes"])
    ssh_command.extend(["-o", "ServerAliveInterval=5"])
    ssh_command.extend(["-o", "ServerAliveCountMax=24"])

    # Exit when forwarding fails
    ssh_command.extend(["-o", "ExitOnForwardFailure=yes"])

    # Work out if we're using our own control master config or not
    has_control_master_config = False
    for ssh_arg in ssh_args:
        if "ControlMaster" in ssh_arg:
            has_control_master_config = True
            break
    # Enable control master so subsequent connections use the same socket
    if not has_control_master_config:
        ssh_command.extend(["-o", "ControlMaster=autoask"])
        ssh_command.extend(["-o", f"ControlPath={temp_dir}/smart-ssh-client-%C"])
        ssh_command.extend(["-o", "ControlPersist=no"])

    # Work out if we're using our own strict host key checking or not
    has_strict_host_key_checking = False
    for ssh_arg in ssh_args:
        if "StrictHostKeyChecking" in ssh_arg:
            has_strict_host_key_checking = True
            break
    # Enable strict host key checking if not already set
    if not has_strict_host_key_checking:
        ssh_command.extend(["-o", "StrictHostKeyChecking=accept-new"])

    # Enable compression
    ssh_command.extend(["-o", "Compression=yes"])

    # Set connection timeout to 10s
    ssh_command.extend(["-o", "ConnectTimeout=30"])

    # If we have a port, add it as a config option
    if port:
        ssh_command.extend(["-o", f"Port={port}"])
    # If we have a username, add it as a config option
    if username:
        ssh_command.extend(["-o", f"User={username}"])

    # Re-run oureslves in proxy mode
    ssh_command.extend(
        [
            "-o",
            f"ProxyCommand={sys.argv[0]} "
            "--smart-ssh-host=%h "
            "--smart-ssh-port=%p "
            + (f"--smart-ssh-knock={port_knock} " if port_knock else "")
            + ("-4 " if "-4" in ssh_args else "")
            + ("-6 " if "-6" in ssh_args else ""),
        ],
    )
    # Pass FD to SSH
    ssh_command.extend(["-o", "ProxyUseFDPass=yes"])
    ssh_command.extend(ssh_args)

    ssh_env = {
        "_SMART_SSH_PROXY_MODE": "1",
    }
    # Set default for SSH_ASKPASS_REQUIRE
    if "SSH_ASKPASS_REQUIRE" not in os.environ:
        ssh_env["SSH_ASKPASS_REQUIRE"] = "prefer"
    # Filter only ENV's we need
    for env_key in os.environ:
        if env_key.startswith(("SSH_", "QT_", "XDG_")) or env_key in [
            "DISPLAY",
            "LANG",
            "TERM",
            "XAUTHORITY",
            # We need a few environment variables to make ksshaskpass and others work
            "DESKTOP_SESSION",
            "HOME",
            "USER",
            "WAYLAND_DISPLAY",
        ]:
            ssh_env[env_key] = os.environ[env_key]

    return_code = 1
    try:
        with subprocess.Popen(  # noqa: S603
            ssh_command,
            shell=False,
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
            env=ssh_env,
        ) as ssh_proc:
            # Wait for the ssh process to exit
            ssh_proc.wait()
            return_code = ssh_proc.returncode

    except subprocess.CalledProcessError as e:
        raise SmartSSHClientError(f"Error running 'ssh': {e}") from None

    except KeyboardInterrupt:
        pass

    return return_code


def entrypoint_ssh() -> None:
    """Entry function for ssh."""
    main("ssh")


def entrypoint_scp() -> None:
    """Entry function for scp."""
    main("scp")


def entrypoint_sftp() -> None:
    """Entry function for sftp."""
    main("sftp")


def main(command: str = "ssh") -> None:  # noqa: C901, PLR0912, PLR0915
    """Entry function from the commandline."""

    ssh_args: list[str] = []

    # If we're running in proxy mode, don't output our banner
    if "_SMART_SSH_PROXY_MODE" not in os.environ:
        print(f"Smart SSH Client v{__version__} - Copyright (c) 2023-2024, AllWorldIT", file=sys.stderr)  # noqa: T201

    # Loop with args and split off what we need
    arg = 1
    while arg < len(sys.argv):
        # Skip if we've hit the end of the args, don't use anything past this ourselves
        if sys.argv[arg] == "--":
            # Add rest of arguments to ssh
            ssh_args.extend(sys.argv[arg:])
            break
        # Check if this is one of our own args, if it is, add it to our arg list
        if sys.argv[arg] in ["--smart-ssh-disable-bitwarden", "--smart-ssh-bitwarden-cache-only"]:
            arg += 1
            continue
        # Check if this is one of our own args, if it is, add it to our arg list
        if sys.argv[arg].startswith("--smart-ssh"):
            # Skip over next arg if the value is not part of this one
            if "=" not in sys.argv[arg]:
                arg += 1
            arg += 1
            continue
        # Add arg to both lists
        ssh_args.append(sys.argv[arg])
        arg += 1

    # Parse our args
    parser = argparse.ArgumentParser(pathlib.Path(sys.argv[0]).name, description="Smart SSH Client")

    smartssh_parser = parser.add_argument_group("smart-ssh")
    smartssh_parser.add_argument(
        "-4",
        action="store_true",
        required=False,
        help="Forces use of IPv4 addresses only.",
        dest="smart_ssh_use_ipv4_only",
    )
    smartssh_parser.add_argument(
        "-6",
        action="store_true",
        required=False,
        help="Forces use of IPv6 addresses only.",
        dest="smart_ssh_use_ipv6_only",
    )
    smartssh_parser.add_argument(
        "--smart-ssh-knock",
        metavar="KNOCK_PORT",
        type=str,
        required=False,
        help="Port knocking port number.",
    )
    smartssh_parser.add_argument(
        "--smart-ssh-disable-bitwarden",
        action="store_true",
        required=False,
        help="Disable Bitwarden entirely.",
        dest="smart_ssh_disable_bitwarden",
    )
    smartssh_parser.add_argument(
        "--smart-ssh-bitwarden-cache-only",
        action="store_true",
        required=False,
        help="Only use the Bitwarden cache when retriving host details.",
        dest="smart_ssh_bitwarden_cache_only",
    )

    # If we are in proxy mode, we need a ssh host and a ssh port
    if "_SMART_SSH_PROXY_MODE" in os.environ:
        smartssh_parser.add_argument(
            "--smart-ssh-host",
            metavar="SSH_HOST",
            type=str,
            required=True,
            help="SSH host to proxy to.",
        )
        smartssh_parser.add_argument(
            "--smart-ssh-port",
            metavar="SSH_PORT",
            type=int,
            required=True,
            help="SSH port to proxy to.",
        )

    # Add the parser for the SSH command being run
    elif command == "ssh":
        add_ssh_parser(parser, ssh_args)
    elif command == "scp":
        add_scp_parser(parser, ssh_args)
    elif command == "sftp":
        add_sftp_parser(parser, ssh_args)

    args = parser.parse_args(sys.argv[1:])

    args_dict = vars(args)

    # You cannot use a jump host when using a SSH proxy
    if args_dict.get("ssh_J") or args_dict.get("scp_J") or args_dict.get("sftp_J"):
        print("ERROR: Smart SSH Client - The -J option not supported", file=sys.stderr)  # noqa: T201
        sys.exit(1)

    # Check for necessary environment variables, else run via ourselves as proxy
    if "_SMART_SSH_PROXY_MODE" not in os.environ:
        try:
            sys.exit(run_via_proxy(command, args_dict, ssh_args))
        except SmartSSHClientError as e:
            print(f"ERROR: Smart SSH Client - {e}", file=sys.stderr)  # noqa: T201
            sys.exit(1)

    # Check if we're using a specific family
    only_family: int | None = None
    if args.smart_ssh_use_ipv4_only:
        only_family = socket.AF_INET
    elif args.smart_ssh_use_ipv6_only:
        only_family = socket.AF_INET6

    # Start off with no port knocking
    port_knock = None
    if args.smart_ssh_knock:
        port_knock = args.smart_ssh_knock

    sys.exit(run_proxy(args.smart_ssh_host, args.smart_ssh_port, port_knock, only_family))
