#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2023-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Smart SSH Client configuration."""

import functools
import json
import os
import pathlib
import subprocess
import sys
import time
import tomllib as toml
from typing import Any

from .exceptions import SmartSSHClientError

try:
    import gnupg
except ImportError:
    gnupg = None

__all__ = ["get_host_config"]

CONFIG_FILE = "~/.config/smart-ssh-client/config.toml"
CONFIG_HOSTS_DIR = "~/.config/smart-ssh-client/hosts.d"
CACHE_BITWARDEN_DIR = "~/.cache/smart-ssh-client/bitwarden"


def _get_config(config_file: pathlib.Path) -> dict[str, Any]:
    """Get configuration from a config file."""

    # Expand config filename
    config_file_expanded = pathlib.Path(config_file).expanduser()

    # Parse into dict
    config = {}
    if config_file_expanded.exists():
        config_data = config_file_expanded.read_text(encoding="UTF-8")
        try:
            config = toml.loads(config_data)
        except toml.TOMLDecodeError as e:
            raise SmartSSHClientError(f"Error parsing config file '{config_file}': {e}") from None

    return config


@functools.cache
def get_config() -> dict[str, str]:
    """Get smart-ssh main configuration."""
    return _get_config(pathlib.Path(CONFIG_FILE).expanduser())


def find_config_files_walk(config_directory: pathlib.Path) -> list[pathlib.Path]:
    """Find all config files in a directory."""
    target_extensions = (".ini", ".ini.gpg")
    found_files: list[pathlib.Path] = []

    for root, _, files in os.walk(config_directory, followlinks=True):
        found_files.extend([pathlib.Path(root).joinpath(x) for x in files if x.endswith(target_extensions)])

    return sorted(found_files)


def find_host_config(  # noqa: C901, PLR0912
    config_file: pathlib.Path, hostname: str, port: str | None
) -> tuple[str | None, str | None, dict[str, str]]:
    """Get host configuration."""

    host_config: dict[str, str] = {}

    config_file_expanded = config_file.expanduser()

    # If the file doesn't exist, just return an empty dict
    if not config_file_expanded.exists():
        return None, None, host_config

    # File format
    file_format: str | None = None
    config_file_name = config_file_expanded.name

    decrypted_data = None
    # Try encrypted config file
    if config_file_name.endswith(".gpg"):
        # Check if we have gnupg loaded, if not show an error
        if not gnupg:
            raise SmartSSHClientError("Encrypted config file found, but python-gnupg is not installed")  # noqa: EM101
        # Initialize GnuPG
        gpg = gnupg.GPG()
        # Open encrypted hosts file
        with config_file_expanded.open("rb") as encrypted_file:
            encrypted_data = encrypted_file.read()
        # Decrypt the data in memory
        decrypted_data = gpg.decrypt(encrypted_data).data.decode("UTF-8")
        config_file_name = config_file_name[:-4]
    # Try unencrypted config file
    elif config_file.name.endswith((".toml", ".json")):
        with config_file_expanded.open(encoding="UTF-8") as hosts_file:
            decrypted_data = hosts_file.read()

    # If we don't have any decrypted data here we can just return
    if not decrypted_data:
        return None, None, host_config

    # Work out file format
    if config_file_name.endswith(".toml"):
        file_format = "toml"
    elif config_file_name.endswith(".json"):
        file_format = "json"

    # Parse decrypted file
    if file_format == "toml":
        try:
            config = toml.loads(decrypted_data)
        except toml.TOMLDecodeError as e:
            raise SmartSSHClientError(f"Error parsing config file '{config_file}': {e}") from None
        # Check if we have a host defined
        host_config_name = None
        # We first check by hostname:port
        if port and config.get(f"{hostname}:{port}"):
            host_config_name = f"{hostname}:{port}"
        # Then just by hostname
        elif config.get(hostname):
            host_config_name = hostname
        # If we found a host key, return the config
        if host_config_name:
            # Set host config
            return f"{config_file}", host_config_name, config[host_config_name]

    elif file_format == "json":
        try:
            config = json.loads(decrypted_data)
        except json.JSONDecodeError as e:
            raise SmartSSHClientError(f"Error parsing config file '{config_file}': {e}") from None
        # Make sure config is a dict
        if not isinstance(config, dict):
            raise SmartSSHClientError(f"Error parsing config file '{config_file}': Not a dictionary")
        # Check if we have a host defined
        host_config_name = None
        # We first check by hostname:port
        if port and config.get(f"{hostname}:{port}"):
            host_config_name = f"{hostname}:{port}"
        # Then just by hostname
        elif config.get(hostname):
            host_config_name = hostname
        # If we found a host key, return the config
        if host_config_name:
            # Set host config
            return f"{config_file}", host_config_name, config[host_config_name]

    return None, None, {}


def get_host_config(args_dict: dict[str, Any], hostname: str, port: str | None) -> dict[str, str]:  # noqa: C901, PLR0912, PLR0915
    """Get host configuration."""

    host_config: dict[str, str] = {}

    # Grab a list of config files
    config_file_list = find_config_files_walk(pathlib.Path(CONFIG_HOSTS_DIR).expanduser())

    # Loop with all config files
    for config_file in config_file_list:
        host_config_file, host_config_name, host_config = find_host_config(config_file, hostname, port)
        if host_config:
            print(  # noqa: T201
                f"INFO: Smart SSH Client - Loaded host '{host_config_name}' configuration from '{host_config_file}'",
                file=sys.stderr,
            )
            break

    # This is the Bitwarden section
    if not host_config:
        sssh_config: dict[str, Any] = get_config()
        # Check if we're using the Bitwarden CLI to get configuration info
        use_bitwarden = sssh_config.get("bitwarden", {}).get("enable", False)
        # If we don't use Bitwarden, just return the host config (blank)
        if not use_bitwarden or args_dict["smart_ssh_disable_bitwarden"]:
            return host_config
        # Set up some variables we need
        sssh_hostname = f"ssh://{hostname}"
        # If we have a port add it to the cache filename
        bwcache_dir = pathlib.Path(CACHE_BITWARDEN_DIR).expanduser()
        bwcache_file = bwcache_dir.joinpath(f"{hostname}:{port}.json" if port else f"{hostname}.json")
        # Check if we have a bitwarden host config file and its modified within 3 months
        if bwcache_file.exists() and (
            bwcache_file.stat().st_mtime > time.time() - sssh_config.get("bitwarden", {}).get("cache_time", 7776000)
            or args_dict["smart_ssh_bitwarden_cache_only"]
        ):
            _, bwcache_hostname, host_config = find_host_config(bwcache_file, hostname, port)
            print(  # noqa: T201
                f"INFO: Smart SSH Client - Using Bitwarden CLI cached data for '{bwcache_hostname}", file=sys.stderr
            )
            return host_config
        # Check if we're only using the cache...
        if args_dict["smart_ssh_bitwarden_cache_only"]:
            return host_config
        # If we do use Bitwarden, try to get the host config
        print("INFO: Smart SSH Client - Trying Bitwarden CLI...", file=sys.stderr)  # noqa: T201
        # Check if we have the Bitwarden CLI installed
        if not pathlib.Path("/usr/bin/bw").exists():
            raise SmartSSHClientError("Bitwarden CLI enabled but not found at '/usr/bin/bw'")  # noqa: EM101
        # Get the Bitwarden CLI data
        bwcli_args = ["/usr/bin/bw", "list", "items", "--url", sssh_hostname]
        with subprocess.Popen(bwcli_args, stdout=subprocess.PIPE) as bwcli:  # noqa: S603
            out, err = bwcli.communicate()
            # Check return value
            rc = bwcli.wait()
        if rc:
            raise SmartSSHClientError(f"Bitwarden CLI failed with return code {rc}: {err!r}")
        # Try parse the JSON output
        try:
            bwcli_rawdata = json.loads(out)
        except json.JSONDecodeError as e:
            raise SmartSSHClientError(f"Error parsing Bitwarden CLI output: {e}") from None
        # Filter hosts that match what we're looking for
        bwcli_hosts: list[Any] = []
        # Loop with each host returned from the search
        if port:
            for bwcli_host in bwcli_rawdata:
                # Loop with login URI's
                for bwcli_host_uri in bwcli_host.get("login", {}).get("uris", []):
                    # We're only interested in SSH URI's
                    if not bwcli_host_uri.get("uri", "").startswith("ssh://"):
                        continue
                    # Make sure it's an exact match
                    if bwcli_host_uri.get("match", -1) != 3:  # noqa: PLR2004
                        continue
                    # Check if the URI matches the hostname
                    if bwcli_host_uri.get("uri", "") != f"{sssh_hostname}:{port}":
                        continue
                    bwcli_hosts.append(bwcli_host)
        # If we didn't get anything with the port try without it
        if not bwcli_hosts:
            # Loop with each host returned from the search and check without port
            for bwcli_host in bwcli_rawdata:
                # Loop with login URI's
                for bwcli_host_uri in bwcli_host.get("login", {}).get("uris", []):
                    # We're only interested in SSH URI's
                    if not bwcli_host_uri.get("uri", "").startswith("ssh://"):
                        continue
                    # Make sure it's an exact match
                    if bwcli_host_uri.get("match", -1) != 3:  # noqa: PLR2004
                        continue
                    # Check if the URI matches the hostname
                    if bwcli_host_uri.get("uri", "") != sssh_hostname:
                        continue
                    bwcli_hosts.append(bwcli_host)
        # Make sure we got results
        if not bwcli_hosts:
            return host_config
        # Make sure we don't have more than one result
        if len(bwcli_hosts) > 1:
            hosts = [x["name"] for x in bwcli_hosts]
            raise SmartSSHClientError(f"Bitwarden CLI returned more than one result for '{hostname}': {hosts}")
        # Grab the host data variable
        bwhost_data = bwcli_hosts[0]
        # Output that we're using the bitwarden host config
        print(  # noqa: T201
            f"INFO: Smart SSH Client - Using Bitwarden CLI host configuration for '{bwhost_data['name']}'", file=sys.stderr
        )
        # Check for supported fields
        if "fields" not in bwhost_data:
            raise SmartSSHClientError(
                f"Bitwarden CLI host '{bwhost_data['name']}' has no 'fields' data, supported options include: "
                "'port', 'portknock', 'username'",
            )
        # Parse the fields which has a 'name' and 'value' key and 'type' of 0
        for field in bwhost_data["fields"]:
            # Make sure we're only using type 0 fields
            if field.get("type", -1) != 0:
                continue
            # Check if we have a overriding hostname
            if field["name"] == "hostname":
                # Set the field value
                host_config[field["name"]] = field["value"]
            # Check if we have a port and validate it
            if field["name"] == "port":
                try:
                    int(field["value"])
                except ValueError:
                    raise SmartSSHClientError(
                        f"Bitwarden CLI host '{bwhost_data['name']}' has an invalid 'port': {field['value']}"
                    ) from None
                # Set the field value
                host_config[field["name"]] = field["value"]
            # Chck if we have a portknock port and validate it
            elif field["name"] == "portknock":
                for pkport in field["value"].split(","):
                    try:
                        int(pkport)
                    except ValueError:
                        raise SmartSSHClientError(
                            f"Bitwarden CLI host '{bwhost_data['name']}' has an invalid 'portknock' port: {port}"
                        ) from None
                # Set the field value
                host_config[field["name"]] = field["value"]
        # Check if we have a username
        if bwhost_data.get("login", {}).get("username"):
            # Set the field value
            host_config["username"] = bwhost_data["login"]["username"]
        # Cache host config
        bwcache_file.parent.mkdir(parents=True, exist_ok=True)
        # Set cache section name
        bwcache_name = f"{hostname}:{port}" if port else hostname
        # Write out cache
        bwcache_file.write_text(json.dumps({bwcache_name: host_config}), encoding="UTF-8")

    return host_config
