#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2023-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""SSH specific functions."""

import argparse
import os

__all__ = [
    "parse_ssh_format",
    "add_ssh_parser",
    "add_scp_parser",
    "add_sftp_parser",
]


def parse_ssh_format(
    ssh_string: str,
) -> tuple[str | None, str | None, str | None]:
    """Parse SSH string into username, host, and port."""

    # Default values
    username: str | None
    host: str | None
    port: str | None

    username, host, port = os.environ.get("USER"), None, None

    # Splitting the username and the rest (if username is provided)
    if "@" in ssh_string:
        username, ssh_string = ssh_string.split("@", 1)

    # Splitting the host and port
    if ":" in ssh_string:
        host, port = ssh_string.split(":", 1)
    else:
        host = ssh_string

    return username, host, port


def add_ssh_parser(  # noqa: PLR0915
    subparser: argparse.ArgumentParser,
    args: list[str],
) -> None:
    """Add the SSH subparser options."""

    # SSH options
    ssh_parser = subparser.add_argument_group("ssh")

    ssh_parser.add_argument(
        "-2",
        action="store_true",
        required=False,
        help="Forces ssh to try protocol version 2 only.",
        dest="ssh_2",
    )

    ssh_parser.add_argument(
        "-A",
        action="store_true",
        required=False,
        help="Enables forwarding of connections from an authentication agent.",
        dest="ssh_A",
    )
    ssh_parser.add_argument(
        "-a",
        action="store_true",
        required=False,
        help="Disables forwarding of connections from an authentication agent.",
        dest="ssh_a",
    )
    ssh_parser.add_argument(
        "-B",
        metavar="BIND_INTERFACE",
        type=str,
        help="Bind to the address of BIND_INTERFACE before attempting to connect to the destination host.",
        dest="ssh_B",
    )
    ssh_parser.add_argument(
        "-b",
        metavar="BIND_INTERFACE",
        type=str,
        help="Use BIND_INTERFACE on the local machine as the source address of the connection.",
        dest="ssh_b",
    )
    ssh_parser.add_argument(
        "-C",
        action="store_true",
        required=False,
        help="Requests  compression  of all data (including stdin, stdout, stderr, and data for forwarded X11, TCP and "
        "Unix-domain connections).",
        dest="ssh_C",
    )
    ssh_parser.add_argument(
        "-c",
        metavar="CIPHER_SPEC",
        type=str,
        help="Selects the cipher specification for encrypting the session.",
        dest="ssh_c",
    )
    ssh_parser.add_argument(
        "-D",
        metavar="[BIND_ADDRESS:]PORT",
        type=str,
        help="Specifies a local “dynamic” application-level port forwarding.",
        dest="ssh_D",
    )
    ssh_parser.add_argument(
        "-E",
        metavar="LOG_FILE",
        type=str,
        help="Append debug logs to LOG_FILE instead of standard error.",
        dest="ssh_E",
    )
    ssh_parser.add_argument(
        "-e",
        metavar="ESCAPE_CHAR",
        type=str,
        help="Sets the escape character for sessions with a pty (default: '~').",
        dest="ssh_e",
    )
    ssh_parser.add_argument(
        "-F",
        metavar="CONFIG_FILE",
        type=str,
        help="Specifies an alternative per-user configuration file.",
        dest="ssh_F",
    )
    ssh_parser.add_argument(
        "-f",
        action="store_true",
        required=False,
        help="Requests ssh to go to background just before command execution.",
        dest="ssh_f",
    )
    ssh_parser.add_argument(
        "-G",
        action="store_true",
        required=False,
        help="Causes ssh to print its configuration after evaluating Host and Match blocks and exit.",
        dest="ssh_G",
    )
    ssh_parser.add_argument(
        "-g",
        action="store_true",
        required=False,
        help="Allows remote hosts to connect to local forwarded ports.",
        dest="ssh_g",
    )
    ssh_parser.add_argument(
        "-I",
        metavar="PKCS11",
        type=str,
        help="Specify the PKCS#11 shared library ssh should use to communicate with a PKCS#11 token providing keys for user "
        "authentication.",
        dest="ssh_I",
    )
    ssh_parser.add_argument(
        "-i",
        metavar="IDENTITY_FILE",
        type=str,
        help="Selects a file from which the identity (private key) for public key authentication is read.",
        dest="ssh_i",
    )
    ssh_parser.add_argument(
        "-J",
        metavar="JUMP_HOST",
        type=str,
        help="Connect to the TARGET via the JUMP_HOST.",
        dest="ssh_J",
    )
    ssh_parser.add_argument(
        "-K",
        action="store_true",
        required=False,
        help="Enables GSSAPI-based authentication and forwarding (delegation) of GSSAPI credentials to the server.",
        dest="ssh_K",
    )
    ssh_parser.add_argument(
        "-k",
        action="store_true",
        required=False,
        help="Disables forwarding (delegation) of GSSAPI credentials to the server.",
        dest="ssh_k",
    )
    ssh_parser.add_argument(
        "-L",
        metavar="FORWARD_SPEC",
        type=str,
        help="Specifies that connections are to be forwarded to the given host and port.",
        dest="ssh_L",
    )
    ssh_parser.add_argument(
        "-l",
        metavar="LOGIN_NAME",
        type=str,
        help="Specifies the user to log in as on the remote machine.",
        dest="ssh_l",
    )
    ssh_parser.add_argument(
        "-M",
        action="store_true",
        required=False,
        help="Places the ssh client into 'master mode for connection sharing.",
        dest="ssh_M",
    )
    ssh_parser.add_argument(
        "-m",
        metavar="MAC_SEC",
        type=str,
        help="A comma-separated list of MAC (message authentication code) algorithms, specified in order of preference.",
        dest="ssh_m",
    )
    ssh_parser.add_argument(
        "-N",
        action="store_true",
        required=False,
        help="Do not execute a remote command.",
        dest="ssh_N",
    )
    ssh_parser.add_argument(
        "-n",
        action="store_true",
        required=False,
        help="Redirects stdin from /dev/null (prevents reading from stdin).",
        dest="ssh_n",
    )
    ssh_parser.add_argument(
        "-O",
        metavar="CTL_CMD",
        type=str,
        help="Control an active connection multiplexing master process.",
        dest="ssh_O",
    )
    ssh_parser.add_argument(
        "-o",
        metavar="OPTION",
        type=str,
        help="Can be used to give options in the format used in the configuration file.",
        dest="ssh_o",
    )
    ssh_parser.add_argument(
        "-P",
        metavar="TAG",
        type=str,
        help="Specify a tag name that may be used to select configuration in ssh_config(5).",
        dest="ssh_P",
    )
    ssh_parser.add_argument(
        "-p",
        metavar="PORT",
        type=int,
        help="Port to connect to on the remote host.",
        dest="ssh_p",
    )
    ssh_parser.add_argument(
        "-Q",
        metavar="QUERY_OPTION",
        type=str,
        help="Queries for the algorithms supported.",
        dest="ssh_Q",
    )
    ssh_parser.add_argument(
        "-q",
        action="store_true",
        required=False,
        help="Quiet mode.",
        dest="ssh_q",
    )
    ssh_parser.add_argument(
        "-R",
        metavar="FORWARD_SPEC",
        type=str,
        help="Specifies that connections are to be forwarded locally the given host and port (reverse port forwarding).",
        dest="ssh_R",
    )
    ssh_parser.add_argument(
        "-S",
        metavar="CTL_PATH",
        type=str,
        help="Specifies the location of a control socket for connection sharing, or the string 'none' to disable connection "
        "sharing.",
        dest="ssh_S",
    )
    ssh_parser.add_argument(
        "-s",
        action="store_true",
        required=False,
        help="May be used to request invocation of a subsystem on the remote system.",
        dest="ssh_s",
    )
    ssh_parser.add_argument(
        "-T",
        action="store_true",
        required=False,
        help="Disable pseudo-terminal allocation.",
        dest="ssh_T",
    )
    ssh_parser.add_argument(
        "-t",
        action="store_true",
        required=False,
        help="Force pseudo-terminal allocation.",
        dest="ssh_t",
    )
    ssh_parser.add_argument(
        "-V",
        action="store_true",
        required=False,
        help="Display the version number and exit.",
        dest="ssh_V",
    )
    ssh_parser.add_argument(
        "-v",
        action="count",
        required=False,
        help="Verbose mode.",
        dest="ssh_v",
    )
    ssh_parser.add_argument(
        "-W",
        metavar="HOST:PORT",
        type=str,
        help="Requests that standard input and output on the client be forwarded to host on port over the secure channel.",
        dest="ssh_W",
    )
    ssh_parser.add_argument(
        "-w",
        metavar="LOCAL_TUN[:REMOTE_TUN]",
        type=str,
        help="Requests tunnel device forwarding with the specified tun devices between the client and the server.",
        dest="ssh_w",
    )
    ssh_parser.add_argument(
        "-X",
        action="store_true",
        required=False,
        help="Enables X11 forwarding.",
        dest="ssh_X",
    )
    ssh_parser.add_argument(
        "-x",
        action="store_true",
        required=False,
        help="Disables X11 forwarding.",
        dest="ssh_x",
    )
    ssh_parser.add_argument(
        "-Y",
        action="store_true",
        required=False,
        help="Enables trusted X11 forwarding.",
        dest="ssh_Y",
    )
    ssh_parser.add_argument(
        "-y",
        action="store_true",
        required=False,
        help="Send log information using syslog.",
        dest="ssh_y",
    )

    needs_target = True
    for ssh_arg in args:
        if ssh_arg in ["-h", "-Q", "-V"]:
            needs_target = False
            break
    if needs_target:
        ssh_parser.add_argument(
            "DESTINATION",
            help="Destination host.",
        )
        ssh_parser.add_argument(
            "COMMAND ARGUMENTS...",
            nargs=argparse.REMAINDER,
            help="Command to run on target host.",
        )


def add_scp_parser(subparser: argparse.ArgumentParser, args: list[str]) -> None:
    """Add the SCP subparser options."""

    # SCP options
    scp_parser = subparser.add_argument_group("scp")
    scp_parser.add_argument(
        "-3",
        action="store_true",
        required=False,
        help="Copies between two remote hosts are transferred through the local host.",
        dest="scp_3",
    )
    scp_parser.add_argument(
        "-A",
        action="store_true",
        required=False,
        help="Enables forwarding of connections from an authentication agent.",
        dest="scp_A",
    )
    scp_parser.add_argument(
        "-B",
        action="store_true",
        required=False,
        help="Selects batch mode.",
        dest="scp_B",
    )
    scp_parser.add_argument(
        "-C",
        action="store_true",
        required=False,
        help="Compression enable.",
        dest="scp_C",
    )
    scp_parser.add_argument(
        "-c",
        metavar="CIPHER_SPEC",
        type=str,
        help="Selects the cipher to use for encrypting the data transfer.",
        dest="ssh_c",
    )
    scp_parser.add_argument(
        "-D",
        metavar="SFTP_SERVER_PATH",
        type=str,
        help="Connect directly to a local SFTP server program rather than a remote one via ssh.",
        dest="scp_D",
    )
    scp_parser.add_argument(
        "-F",
        metavar="CONFIG_FILE",
        type=str,
        help="Specifies an alternative per-user configuration file for ssh.",
        dest="scp_F",
    )
    scp_parser.add_argument(
        "-i",
        metavar="IDENTITY_FILE",
        type=str,
        help="Selects the file from which the identity (private key) for public key authentication is read.",
        dest="scp_i",
    )
    scp_parser.add_argument(
        "-J",
        metavar="JUMP_HOST",
        type=str,
        help="Connect to the TARGET via the JUMP_HOST.",
        dest="scp_J",
    )
    scp_parser.add_argument(
        "-l",
        metavar="BANDWIDTH_LIMIT",
        type=int,
        help="Limit bandwidth in Kbit/s.",
        dest="scp_l",
    )
    scp_parser.add_argument(
        "-O",
        action="store_true",
        required=False,
        help="Use the legacy SCP protocol for file transfers instead of the SFTP protocol.",
        dest="scp_O",
    )
    scp_parser.add_argument(
        "-o",
        metavar="OPTION",
        type=str,
        help="Can be used to pass options to ssh in the format used in ssh_config.",
        dest="scp_o",
    )
    scp_parser.add_argument(
        "-P",
        metavar="PORT",
        type=int,
        help="Specifies the port to connect to on the remote host.",
        dest="scp_P",
    )
    scp_parser.add_argument(
        "-p",
        action="store_true",
        required=False,
        help="Preserves modification times, access times, and file mode bits from the source file.",
        dest="scp_p",
    )
    scp_parser.add_argument(
        "-q",
        action="store_true",
        required=False,
        help="Quiet mode: disables the progress meter as well as warning and diagnostic messages from ssh.",
        dest="scp_q",
    )
    scp_parser.add_argument(
        "-R",
        action="store_true",
        required=False,
        help="Copies between two remote hosts are performed by connecting to the origin host and executing scp there.",
        dest="scp_R",
    )
    scp_parser.add_argument(
        "-r",
        action="store_true",
        required=False,
        help="Recursively copy entire directories.",
        dest="scp_r",
    )
    scp_parser.add_argument(
        "-S",
        metavar="PROGRAM",
        type=str,
        help="Name of program to use for the encrypted connection.",
        dest="scp_S",
    )
    scp_parser.add_argument(
        "-T",
        action="store_true",
        required=False,
        help="Disable strict filename checking.",
        dest="scp_T",
    )
    scp_parser.add_argument(
        "-v",
        action="count",
        required=False,
        help="Verbose mode.",
        dest="scp_v",
    )
    scp_parser.add_argument(
        "-X",
        metavar="SFTP_OPTION",
        type=str,
        help="Specify an option that controls aspects of SFTP protocol behaviour.",
        dest="scp_X",
    )

    needs_target = True
    for ssh_arg in args:
        if ssh_arg == "-h":
            needs_target = False
            break
    if needs_target:
        scp_parser.add_argument(
            "SOURCES.. DESTINATION",
            nargs="+",
            help="Source(s) and destination.",
        )


def add_sftp_parser(subparser: argparse.ArgumentParser, args: list[str]) -> None:
    """Add the SFTP subparser options."""

    # SFTP options
    sftp_parser = subparser.add_argument_group("sftp")

    sftp_parser.add_argument(
        "-a",
        action="store_true",
        required=False,
        help="Attempt to continue interrupted transfers rather than overwriting existing partial or complete copies of files.",
        dest="sftp_a",
    )
    sftp_parser.add_argument(
        "-B",
        metavar="BUFFER_SIZE",
        type=int,
        help="Specify the size of the buffer that sftp uses when transferring files.",
        dest="sftp_B",
    )
    sftp_parser.add_argument(
        "-b",
        metavar="BATCHFILE",
        type=str,
        help="Batch mode reads a series of commands from an input batchfile instead of stdin.",
        dest="sftp_b",
    )
    sftp_parser.add_argument(
        "-C",
        action="store_true",
        required=False,
        help="Enables compression.",
        dest="sftp_C",
    )
    sftp_parser.add_argument(
        "-D",
        metavar="SFTP_SERVER_COMMAND",
        type=str,
        help="Connect directly to a local sftp server.",
        dest="sftp_D",
    )
    sftp_parser.add_argument(
        "-F",
        metavar="SSH_CONFIG",
        type=str,
        help="Specifies an alternative per-user configuration file.",
        dest="sftp_F",
    )
    sftp_parser.add_argument(
        "-f",
        action="store_true",
        required=False,
        help="Requests that files be flushed to disk immediately after transfer.",
        dest="sftp_f",
    )
    sftp_parser.add_argument(
        "-i",
        metavar="IDENTITY_FILE",
        type=str,
        help="Selects the file from which the identity (private key) for public key authentication is read.",
        dest="sftp_i",
    )
    sftp_parser.add_argument(
        "-J",
        metavar="JUMP_HOST",
        type=str,
        help="Connect to the TARGET via the JUMP_HOST.",
        dest="sftp_J",
    )
    sftp_parser.add_argument(
        "-l",
        metavar="LIMIT",
        type=int,
        help="Limits the used bandwidth, specified in Kbit/s.",
        dest="sftp_l",
    )
    sftp_parser.add_argument(
        "-N",
        action="store_true",
        required=False,
        help="Disables quiet mode.",
        dest="sftp_N",
    )
    sftp_parser.add_argument(
        "-o",
        metavar="OPTION",
        type=str,
        help="Can be used to pass options to ssh in the format used in ssh_config.",
        dest="sftp_o",
    )
    sftp_parser.add_argument(
        "-P",
        metavar="PORT",
        type=int,
        help="Specifies the port to connect to on the remote host.",
        dest="sftp_P",
    )
    sftp_parser.add_argument(
        "-p",
        action="store_true",
        required=False,
        help="Preserves modification times, access times, and modes from the original files transferred.",
        dest="sftp_p",
    )
    sftp_parser.add_argument(
        "-q",
        action="store_true",
        required=False,
        help="Quiet mode.",
        dest="sftp_q",
    )
    sftp_parser.add_argument(
        "-R",
        metavar="NUM_REQUESTS",
        type=int,
        help="Specify  how  many requests may be outstanding at any one time.",
        dest="sftp_R",
    )
    sftp_parser.add_argument(
        "-r",
        action="store_true",
        required=False,
        help="Recursively copy entire directories when uploading and downloading.",
        dest="sftp_r",
    )
    sftp_parser.add_argument(
        "-S",
        metavar="PROGRAM",
        type=str,
        help="Name of the program to use for the encrypted connection.",
        dest="sftp_S",
    )
    sftp_parser.add_argument(
        "-s",
        metavar="SUBSYSTEM",
        type=str,
        help="Specifies the SSH2 subsystem or the path for an sftp server on the remote host.",
        dest="sftp_s",
    )
    sftp_parser.add_argument(
        "-v",
        action="count",
        required=False,
        help="Raise logging level.",
        dest="sftp_v",
    )
    sftp_parser.add_argument(
        "-X",
        metavar="SFTP_OPTION",
        type=str,
        help="Specify an option that controls aspects of SFTP protocol behaviour.",
        dest="sftp_X",
    )

    needs_target = True
    for ssh_arg in args:
        if ssh_arg == "-h":
            needs_target = False
            break
    if needs_target:
        sftp_parser.add_argument(
            "DESTINATION",
            type=str,
            help="Destination.",
        )
