"""Utility functions for the Smart SSH Client."""

import os
import socket

from .exceptions import SmartSSHClientError

__all__ = [
    "resolve_host",
    "get_temp_directory",
]


def resolve_host(hostname: str, only_family: int | None = None) -> tuple[int, str]:
    """Resolve hostname to an IP address with a preference for IPv6."""
    try:
        # Loop with GAI results
        for family, _, _, _, sockaddr in socket.getaddrinfo(hostname, None):
            # Check if we have only_family set
            if only_family and family != only_family:
                continue
            # If we didn't find an IPv6 address, return the first address we found
            return (family, sockaddr[0])

        raise SmartSSHClientError(f"Error resolving hostname '{hostname}': No usable results")

    except OSError as e:
        raise SmartSSHClientError(f"Error resolving hostname '{hostname}': {e}") from None


def get_temp_directory() -> str:
    """Get the temporary directory to use for the proxy socket."""

    # Check for XDG_RUNTIME_DIR, TMP_DIR, and TEMP_DIR in the environment
    # Fall back to '/tmp' if none are set
    return os.environ.get("XDG_RUNTIME_DIR") or os.environ.get("TMP_DIR") or os.environ.get("TEMP_DIR") or "/tmp"  # noqa: S108
